# -*- coding: UTF-8 -*-
import requests
from lxml import html

# The required SEC EDGAR request header
SEC_HEADERS = {
    'user-agent': 'Edgar oit@sec.gov',
    'accept-encoding':  'gzip, deflate',
    'host':  'www.sec.gov',
    'referer': 'https://www.sec.gov/', 
    'cache-control': 'no-cache', 
    #'connection': 'close'
    #'connection': 'keep-alive'
}
# Set new default requests header
headers = SEC_HEADERS

class Document:

  def __init__(self, url, timeout=10):
    self.url = url
    self.text = requests.get(self.url, timeout=timeout, headers=SEC_HEADERS).content

class Documents(str):

  def __get_text_from_list__(self, arr):
    return [val.text_content() for val in arr]

  def __init__(self, url, timeout=10):
    self.url = url
    page = requests.get(self.url, timeout=timeout, headers=SEC_HEADERS)
    tree = html.fromstring(page.content)
    content = tree.find_class("formContent")[0]
    info_head = self.__get_text_from_list__(content.find_class("infoHead"))
    info = self.__get_text_from_list__(content.find_class("info"))
    self.content = dict(zip(info_head, info))
    self.element = html.fromstring(requests.get(self.url, timeout=timeout, headers=SEC_HEADERS).content)

  def __repr__(self):
    return str(self.__dict__)

  def __str__(self):
    return str(self.__dict__)



import unittest
from edgar import Edgar

class TestEdgar(unittest.TestCase):

  def test_split_raw_string_to_cik_name(self):
    input_output = [
      ("591 BEVERAGE, INC.:0001602820:", ("591 BEVERAGE, INC.", "0001602820")),
      ("5:15 FUND, LTD.:0001464817:", ("5:15 FUND, LTD.", "0001464817")),
      ("!J INC:0001438823:", ("!J INC", "0001438823"))
    ]
    for inp, output in input_output:
      self.assertEqual(Edgar.split_raw_string_to_cik_name(inp), output)



from typing import Dict, List
import re
from datetime import datetime
from lxml import etree

def findnth(haystack, needle, n):
  parts= haystack.split(needle, n+1)
  if len(parts)<=n+1:
      return -1
  return len(haystack)-len(parts[-1])-len(needle)

class XBRL(etree.ElementBase):

  def __init__(self, *children, attrib=None, nsmap=None, **_extra):
    super().__init__(*children, attrib=None, nsmap=None, **_extra)
    self.definitions = dict(
        (child.attrib["id"], self.__parse_context__(child)) for child in self.child.getchildren() if not isinstance(child, etree._Comment) and "context" in child.tag)
    self.relevant_children = [child for child in self.child.getchildren() if not isinstance(child, etree._Comment) and "context" not in child.tag]
    children = [child for child in self.child.getchildren() if XBRL.is_parsable(child)]
    for elem in children:
      XBRL.clean_tag(elem)

    self.relevant_children_parsed = children
    self.relevant_children_elements = [XBRLElement(child, context_ref=self.definitions[child.attrib["contextRef"]] if child.attrib.get("contextRef") else None) for child in children]

  def __parse_context__(self, context):
    children = [child for child in context.getchildren() if not isinstance(child, etree._Comment)]
    [XBRL.clean_tag(child) for child in children]
    period = [child for child in children if child.tag == 'period'][0]
    return {
        "period": self.__parse_base_elem__(period)
        }

  def __parse_base_elem__(self, elem):
    children = [child for child in elem.getchildren() if XBRL.is_parsable(child)]
    [XBRL.clean_tag(child) for child in children]
    return dict((child.tag, child.text) for child in children)

  @classmethod
  def is_parsable(cls, child):
    return not isinstance(child, etree._Comment) and "context" not in child.tag and "unit" not in child.tag and "schemaRef" not in child.tag

  @classmethod
  def clean_tag(cls, elem):
    """
    Parse tag so 
      {http://fasb.org/us-gaap/2018-01-31}Assets
    becomes
      Assets
    """
    elem.tag = elem.tag[elem.tag.find("}")+1:]

  @classmethod
  def parse_context_ref(cls, context_ref):
    """
    Duration_1_1_2018_To_12_31_2018 becomes 2018-01-01 to 2018-12-31
    As_Of_12_31_2017 becomes 2017-12-31
    """
    context_ref_to_date_text = lambda s: datetime.strptime(s, "%m_%d_%Y").date().strftime("%Y-%m-%d")
    if context_ref.startswith("Duration"):
      if len(context_ref.split("_")) <= 9:
        from_date = context_ref_to_date_text(context_ref[len("DURATION")+1:context_ref.find("_To_")])
        to_date = context_ref_to_date_text(context_ref[context_ref.find("_To_")+4:])
        return {"from": from_date, "to": to_date}
      else:
        from_date = context_ref_to_date_text(context_ref[len("DURATION")+1:context_ref.find("_To_")])
        end_idx = findnth(context_ref, "_", 7)+1
        to_date = context_ref_to_date_text(context_ref[context_ref.find("_To_")+4:end_idx-1])
        return {"from": from_date, "to": to_date}

    elif context_ref.startswith("As_Of"):
      if len(context_ref.split("_")) <= 5:
        return {"from": context_ref_to_date_text(context_ref[len("As_Of")+1:])}
      else:
        end_idx = findnth(context_ref, "_", 4)+1
        from_date = context_ref_to_date_text(context_ref[len("As_Of")+1:end_idx-1])
        return {"from": from_date}
    else:
      return {"other": context_ref.split("_")[0]}

  @property
  def child(self):
    return self.getchildren()[0]

  def find_relevant_elements_by_name(self, name):
    return [elem for elem in self.relevant_children_elements if name.lower() in elem.name.lower()]

  def match_relevant_elements_by_name(self, name):
    return [elem for elem in self.relevant_children_elements if name.lower() == elem.name.lower()]

class XBRLElement(etree.ElementBase):

  def __init__(self, *children, attrib=None, nsmap=None, context_ref=None, **_extra):
    super().__init__(*children, attrib=None, nsmap=None, **_extra)
    self.child = self.getchildren()[0]
    self.context_ref = context_ref
    self.name = ' '.join(re.findall('[A-Z][^A-Z]*', self.child.tag))
    self.unit_ref = self.attrib.get("unitRef") or None

  @property
  def attrib(self) -> Dict:
    return self.child.attrib

  @property
  def value(self) -> str:
    return self.child.text.replace("\n", "").strip() if self.child.text else ""

  def to_dict(self) -> Dict:
    if self.context_ref and self.context_ref.get("period"):
      return { **{
        "name": self.name,
        "value": self.value,
        "unit_ref": self.unit_ref
      }, **self.context_ref["period"]}
    else:
      return {
        "name": self.name,
        "value": self.value,
        "unit_ref": self.unit_ref
      }

  def __repr__(self):
    return f'<{self.name}="{self.value} {self.unit_ref}" context_ref={self.context_ref}>'



# -*- coding: UTF-8 -*-
from typing import List
import os
import requests
from lxml import html, etree
from .document import Documents
import lxml

BASE_URL = "https://www.sec.gov"

# The required SEC EDGAR request header
SEC_HEADERS = {
    'user-agent': 'Edgar oit@sec.gov',
    'accept-encoding':  'gzip, deflate',
    'host':  'www.sec.gov',
    'referer': 'https://www.sec.gov/', 
    'cache-control': 'no-cache', 
    #'connection': 'close'
    #'connection': 'keep-alive'
}
# Set new default requests header
headers = SEC_HEADERS

class Company():

    def __init__(self, name, cik, timeout=10):
        self.name = name
        self.cik = cik
        self.url = f"https://www.sec.gov/cgi-bin/browse-edgar?action=getcompany&CIK={cik}"
        self.timeout = timeout
        self._document_urls = []
        self.get_company_info()

    @property
    def document_urls(self):
      return list(set(self._document_urls))

    def _get(self, url):
      return requests.get(url, timeout=self.timeout, headers=SEC_HEADERS)

    def get_company_info(self):
        page = html.fromstring(self._get(self.url).content)
        companyInfo = page.xpath("//div[@class='companyInfo']")[0] if page.xpath("//div[@class='companyInfo']") else None
        if companyInfo is not None:
          indentInfo = companyInfo.getchildren()[1]
          self.sic = indentInfo.getchildren()[1].text if len(indentInfo.getchildren()) > 2 else ""
          self.us_state = indentInfo.getchildren()[3].text if len(indentInfo.getchildren()) > 4 else ""

        self._document_urls = [ BASE_URL + elem.attrib["href"]
            for elem in page.xpath("//*[@id='documentsbutton']") if elem.attrib.get("href")]

    def get_filings_url(self, filing_type="", prior_to="", ownership="include", no_of_entries=100) -> str:
        url = self.url + "&type=" + filing_type + "&dateb=" + prior_to + "&owner=" +  ownership + "&count=" + str(no_of_entries)
        return url

    def get_all_filings(self, filing_type="", prior_to="", ownership="include", no_of_entries=100) -> lxml.html.HtmlElement:
      url = self.get_filings_url(filing_type, prior_to, ownership, no_of_entries)
      page = self._get(url)
      return html.fromstring(page.content)

    def _group_document_type(self, tree, document_type):
      result = []
      grouped = []
      for i, elem in enumerate(tree.xpath('//*[@id="seriesDiv"]/table/tr')):
        if i == 0:
          continue
        url = elem.xpath("td")[1].getchildren()[0].attrib["href"]
        grouped.append(url)
        if elem.xpath("td")[0].text == document_type:
          result.append(grouped)
          grouped = []
      return result

    def get_document_type_from_10K(self, document_type, no_of_documents=1) -> List[lxml.html.HtmlElement]:
      tree = self.get_all_filings(filing_type="10-K")
      url_groups = self._group_document_type(tree, "10-K")[:no_of_documents]
      result = []
      for url_group in url_groups:
        for url in url_group:
          url = BASE_URL + url
          self._document_urls.append(url)
          content_page = Company.get_request(url)
          table = content_page.find_class("tableFile")[0]
          for row in table.getchildren():
            if document_type in row.getchildren()[3].text:
              href = row.getchildren()[2].getchildren()[0].attrib["href"]
              href = BASE_URL + href
              href = href.replace("ix?doc=/", "") # required for new iXBRL to HTML
              if len(href.split("/")[-1].split(".")) == 1:
                continue
              doc = Company.get_request(href)
              result.append(doc)
      return result

    def get_data_files_from_10K(self, document_type, no_of_documents=1, isxml=False) -> List[lxml.html.HtmlElement]:
      tree = self.get_all_filings(filing_type="10-K")
      url_groups = self._group_document_type(tree, "10-K")[:no_of_documents]
      result = []
      for url_group in url_groups:
        for url in url_group:
          url = BASE_URL + url
          self._document_urls.append(url)
          content_page = Company.get_request(url)
          tableFile = content_page.find_class("tableFile")
          if len(tableFile) < 2:
            continue
          table = tableFile[1]
          for row in table.getchildren():
            if document_type in row.getchildren()[3].text:
              href = row.getchildren()[2].getchildren()[0].attrib["href"]
              href = BASE_URL + href
              doc = Company.get_request(href, isxml=isxml)
              result.append(doc)
      return result

    @classmethod
    def __get_documents_from_element__(cls, elem, as_documents=False):
      url = BASE_URL + elem.attrib["href"]
      if as_documents:
        return Documents(url)
      else:
        content_page = Company.get_request(url)
        table = content_page.find_class("tableFile")[0]
        last_row = table.getchildren()[-1]
        href = last_row.getchildren()[2].getchildren()[0].attrib["href"]
        href = BASE_URL + href
        return Company.get_request(href)

    def get_10Ks(self, no_of_documents=1, as_documents=False) -> List[lxml.html.HtmlElement]:
      tree = self.get_all_filings(filing_type="10-K")
      elems = tree.xpath('//*[@id="documentsbutton"]')[:no_of_documents]
      result = []
      for elem in elems:
          doc = Company.__get_documents_from_element__(elem, as_documents=as_documents)
          result.append(doc)
      return result

    def get_10K(self) -> List[lxml.html.HtmlElement]:
      return self.get_10Ks(no_of_documents=1)[0]

    @classmethod
    def get_request(cls, href, isxml=False, timeout=10):
        page = requests.get(href, timeout=timeout, headers=SEC_HEADERS)
        if isxml:
          p = etree.XMLParser(huge_tree=True)
          return etree.fromstring(page.content, parser=p)
        else:
          return html.fromstring(page.content)

    @classmethod
    def get_documents(cls, tree: html.HtmlElement, no_of_documents=1, debug=False, as_documents=False) -> List:
        elems = tree.xpath('//*[@id="documentsbutton"]')[:no_of_documents]
        result = []
        for elem in elems:
            filing = Company.__get_documents_from_element__(elem, as_documents=as_documents)
            result.append(filing)

        if len(result) == 1:
            return result[0]
        return result

    @classmethod
    def get_CIK_from_company(cls, company_name):
        tree = cls.get_request("https://www.sec.gov/cgi-bin/browse-edgar?company=" + company_name)
        CIKList = tree.xpath('//*[@id="seriesDiv"]/table/tr[*]/td[1]/a/text()')
        names_list = []
        for elem in tree.xpath('//*[@id="seriesDiv"]/table/tr[*]/td[2]'):
            names_list.append(elem.text_content())
        return list(zip(CIKList, names_list))



# -*- coding: UTF-8 -*-
from typing import Tuple, List, Any, Dict
from lxml import html
from tqdm import tqdm
import os
import requests
from rapidfuzz import process, fuzz


class Edgar():

    def __init__(self, companies_page_path=None):

        # The required SEC EDGAR request header
        SEC_HEADERS = {
            'user-agent': 'Edgar oit@sec.gov',
            'accept-encoding':  'gzip, deflate',
            'host':  'www.sec.gov',
            'referer': 'https://www.sec.gov/', 
            'cache-control': 'no-cache', 
            #'connection': 'close'
            #'connection': 'keep-alive'
        }
        # Set new default requests header
        headers = SEC_HEADERS

        # Add patch from:
        # https://github.com/NetSPI/NetblockTool/issues/3#issuecomment-897138800
        # Here we use the while loop as a poor-mans patch for rate limiting.
        # When that occurs, the first item returns a non-html doc...
        rate_limited = 0
        while not rate_limited:

            all_companies_content : str
            if companies_page_path is not None and os.path.isfile(companies_page_path):
                all_companies_content = open(companies_page_path, encoding="latin-1").read()
            else:
                edgar_url = "https://www.sec.gov/Archives/edgar/cik-lookup-data.txt"
                all_companies_page = requests.get(edgar_url, headers=SEC_HEADERS)
                all_companies_content = all_companies_page.content.decode("latin1")
            all_companies_array = all_companies_content.split("\n")
            
            # Check for rate limiting garbage...
            item_arr = all_companies_array[0].split(":")
            if item_arr[0] != '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http':
                rate_limited = 1
            
            del all_companies_array[-1]
            all_companies_array_rev = []
            for i, item in enumerate(all_companies_array):
                if item == "":
                    continue
                _name, _cik = Edgar.split_raw_string_to_cik_name(item)
                all_companies_array[i] = (_name, _cik)
                all_companies_array_rev.append((_cik, _name))
            self.all_companies_dict = dict(all_companies_array)
            self.all_companies_dict_rev = dict(all_companies_array_rev)

    def get_cik_by_company_name(self, name) -> str:
        return self.all_companies_dict[name]

    def match_company_by_company_name(self, name, top=5, progress=True) -> List[Dict[str, Any]]:
        result = []
        for company, cik in (
            tqdm(self.all_companies_dict.items()) if progress else self.all_companies_dict.items()
        ):
            result.append({"company_name": company, "cik": cik, "score": fuzz.partial_ratio(name, company)})
        return sorted(result, key=lambda row: row["score"], reverse=True)[:top]
	
    def get_company_name_by_cik(self, cik) -> str:
        return self.all_companies_dict_rev[cik]

    def find_company_name(self, words) -> List[str]:
        possible_companies = []
        words = words.lower()
        for company in self.all_companies_dict:
            if all(word in company.lower() for word in words.split(" ")):
                possible_companies.append(company)
        return possible_companies

    @classmethod
    def split_raw_string_to_cik_name(cls, item):
        item_arr = item.split(":")[:-1]
        return ":".join(item_arr[:-1]), item_arr[-1]
        
def test():
    com = Company("Oracle Corp", "0001341439")
    tree = com.get_all_filings(filingType="10-K")
    return Company.get_documents(tree)
